from biolib import api
from biolib._internal.runtime import RuntimeJobData as _RuntimeJobData


def set_main_result_prefix(result_prefix: str) -> None:
    runtime_job_data = _RuntimeJobData.get()
    if not runtime_job_data:
        raise Exception('Unable to load BioLib runtime job data') from None

    api.client.patch(
        data={'result_name_prefix': result_prefix},
        headers={'Job-Auth-Token': runtime_job_data['job_auth_token']},
        path=f"/jobs/{runtime_job_data['job_uuid']}/main_result/",
    )



import json
from biolib.typing_utils import TypedDict, Optional, cast


class RuntimeJobDataDict(TypedDict):
    version: str
    job_requested_machine: str
    job_uuid: str
    job_auth_token: str


class RuntimeJobData:
    _job_data: Optional[RuntimeJobDataDict] = None

    @staticmethod
    def get() -> Optional[RuntimeJobDataDict]:
        if not RuntimeJobData._job_data:
            try:
                with open('/biolib/secrets/biolib_system_secret', mode='r') as file:
                    job_data: RuntimeJobDataDict = json.load(file)
            except BaseException:
                return None

            if not job_data['version'].startswith('1.'):
                raise Exception(f"Unexpected system secret version {job_data['version']} expected 1.x.x")

            RuntimeJobData._job_data = job_data

        return cast(RuntimeJobDataDict, RuntimeJobData._job_data)



from biolib._internal.push_application import push_application as _push_application
from biolib._internal.push_application import set_app_version_as_active as _set_app_version_as_active

from biolib.app import BioLibApp as _BioLibApp

def push_app_version(uri: str, path: str) -> _BioLibApp:
    push_data = _push_application(
        app_uri=uri,
        app_path=path,
        app_version_to_copy_images_from=None,
        is_dev_version=True)
    uri = f'{push_data["app_uri"]}:{push_data["sematic_version"]}'
    return _BioLibApp(uri)

def set_app_version_as_default(app_version: _BioLibApp) -> None:
    app_version_uuid = app_version.version['public_id']
    _set_app_version_as_active(app_version_uuid)

def get_app_version_pytest_plugin(app_version: _BioLibApp):
    try:
        import pytest # type: ignore # pylint: disable=import-outside-toplevel,import-error
    except BaseException:
        raise Exception('Failed to import pytest; please make sure it is installed') from None

    class AppVersionFixturePlugin(object):
        def __init__(self, app_version_ref):
            self.app_version_ref = app_version_ref

        @pytest.fixture(scope='session')
        def app_version(self, request): # pylint: disable=unused-argument
            return self.app_version_ref

    return AppVersionFixturePlugin(app_version)



import re
from io import BufferedIOBase
from biolib.typing_utils import List, Optional, Dict, Union

allowed_sequence_chars = set("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_.")
def find_invalid_sequence_characters(sequence):
    invalid_chars = [char for char in sequence if char not in allowed_sequence_chars]
    return invalid_chars

class SeqUtilRecord:
    def __init__(self,
                 sequence: str,
                 sequence_id: str,
                 description: Optional['str'],
                 properties: Optional[Dict[str, str]] = None):
        self.sequence = sequence
        self.id = sequence_id  # pylint: disable=invalid-name
        self.description = description

        if properties:
            disallowed_pattern = re.compile(r"[=\[\]\n]")
            for key, value in properties.items():
                assert not bool(disallowed_pattern.search(key)), "Key cannot contain characters =[] and newline"
                assert not bool(disallowed_pattern.search(value)), "Value cannot contain characters =[] and newline"
            self.properties = properties
        else:
            self.properties = {}

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} ({self.id})'


class SeqUtil:
    @staticmethod
    def parse_fasta(
        input_file: Union[str, BufferedIOBase] = None,
        default_header: Optional[str] = None,
        allow_any_sequence_characters: bool = False,
        allow_empty_sequence: bool = False,
        file_name: Optional[str] = None,
    ) -> List[SeqUtilRecord]:
        if input_file is None:
            if file_name:
                input_file = file_name
            else:
                raise ValueError("input_file must be a file name (str) or a BufferedIOBase object")
        if isinstance(input_file, str):
            with open(input_file, 'r') as file_handle:
                data = file_handle.read().strip()
        elif isinstance(input_file, BufferedIOBase):
            data = input_file.read().decode('utf-8')
        else:
            raise ValueError("input_file must be a file name (str) or a BufferedIOBase object")
        if not data:
            return []

        if '>' not in data:
            if default_header:
                lines_with_header = []
                for index, line in enumerate(data.split('\n')):
                    index_string = str(index + 1) if index > 0 else ''
                    lines_with_header.append(f'>{default_header}{index_string}\n{line}')

                data = '\n'.join(lines_with_header)
            else:
                raise Exception(f'No header line found in FASTA file "{file_name}"')

        splitted = []
        tmp_data = ""
        for line in data.splitlines():
            if line.startswith(">"):
                if tmp_data:
                    splitted.append(tmp_data)
                tmp_data = line[1:].strip() + '\n'
            else:
                if line.strip():
                    tmp_data += line.strip() + '\n'

        if tmp_data:
            splitted.append(tmp_data)

        parsed_sequences = []
        for sequence_data in splitted:
            sequence_data_splitted = sequence_data.strip().split('\n')
            header_line = sequence_data_splitted[0].split()
            sequence_id = header_line[0]
            description = sequence_data_splitted[0][len(sequence_id):].strip()
            sequence = "".join([seq.strip().upper() for seq in sequence_data_splitted[1:]])

            if not allow_any_sequence_characters:
                invalid_sequence_characters = find_invalid_sequence_characters(sequence)
                if len(invalid_sequence_characters) > 0:
                    raise Exception(
                        f'Error: Invalid character ("{invalid_sequence_characters[0]}") found in sequence {sequence_id}'
                    )
            if not allow_empty_sequence and len(sequence) == 0:
                raise Exception(
                    f'Error: No sequence found for fasta entry {sequence_id}'
                )

            parsed_sequences.append(
                SeqUtilRecord(sequence=sequence, sequence_id=sequence_id, description=description)
            )
        return parsed_sequences

    @staticmethod
    def write_records_to_fasta(file_name: str, records: List[SeqUtilRecord]) -> None:
        with open(file_name, mode='w') as file_handle:
            for record in records:
                optional_description = f' {record.description}' if record.description else ''
                if record.properties:
                    for key, value in record.properties.items():
                        optional_description += f' [{key}={value}]'
                sequence = '\n'.join(record.sequence[i:i + 80] for i in range(0, len(record.sequence), 80))
                file_handle.write(f'>{record.id}{optional_description}\n{sequence}\n')



import collections.abc
import multiprocessing
import os
import socket
import sys

from importlib_metadata import version, PackageNotFoundError

from biolib.typing_utils import Optional
from biolib.utils.seq_util import SeqUtil, SeqUtilRecord
from biolib._internal.http_client import HttpClient
from biolib.biolib_logging import logger_no_user_data, logger
from biolib.typing_utils import Tuple, Iterator
from .multipart_uploader import MultiPartUploader, get_chunk_iterator_from_bytes

# try fetching version, if it fails (usually when in dev), add default
try:
    BIOLIB_PACKAGE_VERSION = version('pybiolib')
except PackageNotFoundError:
    BIOLIB_PACKAGE_VERSION = '0.0.0'

IS_DEV = os.getenv('BIOLIB_DEV', '').upper() == 'TRUE'


def load_base_url_from_env() -> str:
    base_url = os.getenv('BIOLIB_BASE_URL')
    if base_url:
        return base_url.lower().rstrip('/')

    try:
        search_list = []
        with open('/etc/resolv.conf') as file:
            for line in file:
                line_trimmed = line.strip()
                if line_trimmed.startswith('search'):
                    search_list = line_trimmed.split()[1:]
                    logger.debug(f'Found search list: {search_list} when resolving base url.')
                    break

        for search_host in search_list:
            host_to_try = f'biolib.{search_host}'
            try:
                if len(socket.getaddrinfo(host_to_try, 443)) > 0:
                    return f'https://{host_to_try}'.lower()
            except BaseException:  # pylint: disable=broad-except
                pass
    except BaseException:  # pylint: disable=broad-except
        pass

    return 'https://biolib.com'


BIOLIB_BASE_URL: Optional[str] = None
BIOLIB_SITE_HOSTNAME: Optional[str] = None

BIOLIB_CLOUD_BASE_URL = os.getenv('BIOLIB_CLOUD_BASE_URL', '').lower()

BIOLIB_PACKAGE_ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

BIOLIB_CLOUD_ENVIRONMENT = os.getenv('BIOLIB_CLOUD_ENVIRONMENT', '').lower()

BIOLIB_SECRETS_TMPFS_PATH = os.environ.get('BIOLIB_SECRETS_TMPFS_PATH')

IS_RUNNING_IN_CLOUD = BIOLIB_CLOUD_ENVIRONMENT == 'non-enclave'

BASE_URL_IS_PUBLIC_BIOLIB: Optional[bool] = None

# sys.stdout is an instance of OutStream in Jupyter and Colab which does not have .buffer
if not hasattr(sys.stdout, 'buffer'):
    IS_RUNNING_IN_NOTEBOOK = True
else:
    IS_RUNNING_IN_NOTEBOOK = False

STREAM_STDOUT = False

if BIOLIB_CLOUD_ENVIRONMENT and not IS_RUNNING_IN_CLOUD:
    logger_no_user_data.warning((
        'BIOLIB_CLOUD_ENVIRONMENT defined but does not specify the cloud environment correctly. ',
        'The compute node will not act as a cloud compute node'
    ))

ByteRangeTuple = Tuple[int, int]
DownloadChunkInputTuple = Tuple[ByteRangeTuple, str]


def _download_chunk(input_tuple: DownloadChunkInputTuple) -> bytes:
    byte_range, presigned_url = input_tuple
    start, end = byte_range

    response = HttpClient.request(
        url=presigned_url,
        headers={'range': f'bytes={start}-{end}'},
        timeout_in_seconds=300,  # timeout after 5 min
        retries=10,
    )
    logger_no_user_data.debug(f'Returning raw data for part {start}')
    return response.content


class ChunkIterator(collections.abc.Iterator):

    def __init__(self, file_size: int, chunk_size: int, presigned_url: str):
        self._semaphore = multiprocessing.BoundedSemaphore(20)  # support 20 chunks to be processed at once
        self._iterator = self._get_chunk_input_iterator(file_size, chunk_size, presigned_url)

    def __iter__(self):
        return self

    def __next__(self):
        if self._semaphore.acquire(timeout=1800):
            return next(self._iterator)
        else:
            raise Exception('Did not receive work within 30 min.')

    def chunk_completed(self) -> None:
        self._semaphore.release()

    @staticmethod
    def _get_chunk_input_iterator(
            file_size: int,
            chunk_size: int,
            presigned_url: str,
    ) -> Iterator[DownloadChunkInputTuple]:
        for index in range(0, file_size, chunk_size):
            byte_range: ByteRangeTuple = (index, index + chunk_size - 1)
            yield byte_range, presigned_url


def download_presigned_s3_url(presigned_url: str, output_file_path: str) -> None:
    chunk_size = 50_000_000

    response = HttpClient.request(url=presigned_url, headers={'range': 'bytes=0-1'})
    file_size = int(response.headers['Content-Range'].split('/')[1])

    chunk_iterator = ChunkIterator(file_size, chunk_size, presigned_url)

    bytes_written = 0
    # use 16 cores, unless less is available
    process_pool = multiprocessing.Pool(processes=min(16, multiprocessing.cpu_count() - 1))
    try:
        with open(output_file_path, 'ab') as output_file:
            for index, data in enumerate(process_pool.imap(_download_chunk, chunk_iterator)):
                logger_no_user_data.debug(f'Writing part {index} to file...')
                output_file.write(data)

                bytes_written += chunk_size
                approx_progress_percent = min(bytes_written / file_size * 100, 100)
                logger_no_user_data.debug(
                    f'Wrote part {index} of {file_size} to file, '
                    f'the approximate progress is {round(approx_progress_percent, 2)}%'
                )
                chunk_iterator.chunk_completed()
    finally:
        logger_no_user_data.debug('Closing process poll...')
        process_pool.close()
        logger_no_user_data.debug('Process poll closed.')



import biolib.api

class BiolibAccountApi:

    @staticmethod
    def fetch_by_handle(account_handle):
        response = biolib.api.client.get(path=f'/account/{account_handle}')
        return response.json()



import biolib.api
from biolib.biolib_api_client.lfs_types import LargeFileSystemVersion, LargeFileSystem


class BiolibLargeFileSystemApi:

    @staticmethod
    def create(account_uuid: str, name: str) -> LargeFileSystem:
        response = biolib.api.client.post(
            path='/lfs/',
            data={'account_uuid': account_uuid, 'name': name},
        )

        lfs: LargeFileSystem = response.json()
        return lfs

    @staticmethod
    def fetch_version(lfs_version_uuid: str) -> LargeFileSystemVersion:
        response = biolib.api.client.get(
            path=f'/lfs/versions/{lfs_version_uuid}/',
        )

        lfs_version: LargeFileSystemVersion = response.json()
        return lfs_version

    @staticmethod
    def create_version(resource_uuid: str) -> LargeFileSystemVersion:
        response = biolib.api.client.post(
            path='/lfs/versions/',
            data={'resource_uuid': resource_uuid},
        )

        lfs_version: LargeFileSystemVersion = response.json()
        return lfs_version



# pylint: disable=unsubscriptable-object
import base64
import json
import os
import subprocess
import time
from datetime import datetime
from socket import gethostbyname, gethostname

from biolib import utils, api
from biolib.biolib_logging import logger_no_user_data
from biolib.typing_utils import Optional, List, Dict, cast
from biolib.biolib_api_client import BiolibApiClient
from biolib.compute_node.webserver.webserver_types import WebserverConfig, ComputeNodeInfo, ShutdownTimes


def trust_ceritificates(certs_data: List[str]) -> None:
    ca_directory_amazon_linux_2 = '/etc/pki/ca-trust/source/anchors/'

    if not os.path.exists(ca_directory_amazon_linux_2):
        logger_no_user_data.error(f'Certificate path not found at {ca_directory_amazon_linux_2}')
        return

    for idx, cert_data in enumerate(certs_data):
        with open(f'{ca_directory_amazon_linux_2}bl-cert-{idx}.crt', mode='w') as cert_file:
            cert_file.write(cert_data)

    result = subprocess.run(['update-ca-trust'], capture_output=True, check=False)
    logger_no_user_data.debug(result.stdout.decode())
    if result.returncode == 0:
        logger_no_user_data.info('Certificates added successfully!')
    else:
        logger_no_user_data.error(f'Failed to update certificates: {result.stderr.decode()}')


class CloudUtils:
    _webserver_config: Optional[WebserverConfig] = None

    @staticmethod
    def initialize() -> None:
        logger_no_user_data.info('Reporting availability...')
        CloudUtils._report_availability()

    @staticmethod
    def get_webserver_config() -> WebserverConfig:
        if CloudUtils._webserver_config:
            return CloudUtils._webserver_config

        CloudUtils._webserver_config = WebserverConfig(
            compute_node_info=ComputeNodeInfo(
                auth_token=CloudUtils._get_environment_variable_or_fail('BIOLIB_COMPUTE_NODE_AUTH_TOKEN'),
                ip_address=os.environ.get('BIOLIB_COMPUTE_NODE_CUSTOM_IP', default=gethostbyname(gethostname())),
                public_id=CloudUtils._get_environment_variable_or_fail('BIOLIB_COMPUTE_NODE_PUBLIC_ID'),
                pybiolib_version=utils.BIOLIB_PACKAGE_VERSION,
            ),
            base_url=CloudUtils._get_environment_variable_or_fail('BIOLIB_BASE_URL'),
            s3_general_storage_bucket_name=CloudUtils._get_environment_variable_or_fail(
                'BIOLIB_S3_GENERAL_STORAGE_BUCKET_NAME',
            ),
            is_dev=os.environ.get('BIOLIB_DEV') == 'TRUE',
            shutdown_times=ShutdownTimes(
                auto_shutdown_time_in_seconds=CloudUtils._get_environment_variable_as_int(
                    'BIOLIB_CLOUD_AUTO_SHUTDOWN_TIME_IN_SECONDS'
                ),
            )
        )

        return CloudUtils._webserver_config

    @staticmethod
    def deregister(error: Optional[str] = None) -> None:
        if utils.IS_RUNNING_IN_CLOUD:
            config = CloudUtils.get_webserver_config()
            try:
                api.client.post(
                    authenticate=False,
                    path='/jobs/deregister/',
                    data={
                        'auth_token': config['compute_node_info']['auth_token'],
                        'public_id': config['compute_node_info']['public_id'],
                        'error': error,
                    },
                )
            except BaseException as error_object:
                logger_no_user_data.error(f'Failed to deregister got error: {error_object}')
        else:
            logger_no_user_data.error("Not deregistering as environment is not cloud")

    @staticmethod
    def shutdown() -> None:
        if utils.IS_RUNNING_IN_CLOUD:
            logger_no_user_data.debug('Waiting 10 seconds and shutting down...')
            # Sleep for 10 seconds to ensure logs are written
            time.sleep(10)
            logger_no_user_data.debug('Shutting down...')
            try:
                subprocess.run(['sudo', 'shutdown', 'now'], check=True)
            except Exception as error:  # pylint: disable=broad-except
                logger_no_user_data.error(f'Failed to shutdown got error: {error}')
        else:
            logger_no_user_data.error("Not running shutdown as environment is not cloud")

    @staticmethod
    def deregister_and_shutdown() -> None:
        logger_no_user_data.debug('Deregistering and shutting down...')

        CloudUtils.deregister()
        CloudUtils.shutdown()

    @staticmethod
    def finish_cloud_job(cloud_job_id: str, system_exception_code: Optional[int], exit_code: Optional[int]) -> None:
        if not cloud_job_id:
            logger_no_user_data.error(
                'Finish cloud job was called but no cloud job was supplied. '
                f'System exception code: {system_exception_code}'
            )
            return

        logger_no_user_data.debug(
            f'Reporting CloudJob "{cloud_job_id}" as finished with exit code: {exit_code} '
            f'and system exception code: {system_exception_code}'
        )

        config = CloudUtils.get_webserver_config()
        try:
            api.client.post(
                authenticate=False,
                path='/jobs/cloud/finish/',
                retries=100,
                data={
                    'auth_token': config['compute_node_info']['auth_token'],
                    'cloud_job_id': cloud_job_id,
                    'system_exception_code': system_exception_code,
                    'exit_code': exit_code
                },
            )
        except BaseException as error:
            logger_no_user_data.debug(f'Failed to finish CloudJob "{cloud_job_id}" due to: {error}')

    @staticmethod
    def _report_availability() -> None:
        try:
            config = CloudUtils.get_webserver_config()
            compute_node_info = config['compute_node_info']
            api_client = BiolibApiClient.get()
            logger_no_user_data.debug(
                f'Registering with {compute_node_info} to host {api_client.base_url} at {datetime.now()}'
            )
            response = api.client.post(
                authenticate=False,
                path='/jobs/report_available/',
                data=cast(Dict[str, str], compute_node_info),
            )
            if response.status_code != 201:
                raise Exception("Non 201 error code")
            else:
                logger_no_user_data.info("Compute node registered!")
                response_data = response.json()
                logger_no_user_data.info(f"Got data on register: {json.dumps(response_data)}")
                certs = []
                for federation in response_data["federation"]:
                    for cert_b64 in federation["certs_b64"]:
                        certs.append(base64.b64decode(cert_b64).decode())
                trust_ceritificates(certs)

        except Exception as exception:  # pylint: disable=broad-except
            logger_no_user_data.error(f'Shutting down as self register failed due to: {exception}')
            if not utils.IS_DEV:
                CloudUtils.deregister_and_shutdown()

    @staticmethod
    def _get_environment_variable_or_fail(key: str) -> str:
        value = os.environ.get(key)
        # Purposely loose falsy check (instead of `is not None`) as empty string should fail
        if not value:
            raise Exception(f'CloudUtils: Missing environment variable "{key}"')

        return value

    @staticmethod
    def _get_environment_variable_as_int(key: str) -> int:
        return int(CloudUtils._get_environment_variable_or_fail(key))
